/*
 * Copyright (c) 2011 Lucas Baudin <xapantu@gmail.com>
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; see the file COPYING.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

public enum Euclide.CMake.CommandType
{
    NOTHING,
    COMMENT,
    FUNCTION,
    ASSIGNMENT
}

/**
 * Base class for all commands.
 **/
public abstract class Euclide.CMake.Command : GLib.Object
{
    public string command;
    public bool can_provide_vars { construct set; get; default = true; }
    public Gee.HashMap<string, string> vars_defined = null;
    public Gee.HashMap<string, string> vars_needed = null;
    public Gee.HashMap<string, Command> vars_needed_cmd = null;

    public abstract void construct_command();
    public abstract void parse();
    
    construct
    {
        vars_defined = new Gee.HashMap<string, string>();
        vars_needed = new Gee.HashMap<string, string>();
        vars_needed_cmd = new Gee.HashMap<string, Command>();
    }

    public virtual void resolve_vars(Gee.ArrayList<Command> commands)
    {
        foreach(var entry_needed in vars_needed.entries)
        {
            if(!search_variable_for_name(entry_needed, commands))
                ; //debug("Couldn't find value for: %s", entry_needed.key);
        }
    }
    
    bool search_variable_for_name(Gee.Map.Entry<string, string> entry_var, Gee.ArrayList<Command> commands)
    {
        for(int i = commands.size - 1; i >= 0; i--)
        {
            if(commands[i] != this && commands[i].can_provide_vars)
            {
                foreach(var entry in commands[i].vars_defined.entries)
                {
                    if(entry.key == entry_var.key)
                    {
                        entry_var.value = entry.value;
                        vars_needed_cmd[entry.key] = commands[i];
                        return true;
                    }
                }
            }
        }
        return false;
    }

    public void get_pkg_config_ancestors(ref List<PkgCheckModules>? list)
    {
        foreach(var cmd in vars_needed_cmd.values)
        {
            if(cmd is PkgCheckModules)
                list.append(cmd as PkgCheckModules);
            else
            {
                cmd.get_pkg_config_ancestors(ref list);
            }
        }
    }
}

/**
 * A comment command, it doesn't do anything, it just stores the comment
 * text.
 **/
public class Euclide.CMake.Comment : Euclide.CMake.Command
{
    public Comment(string comment)
    {
    	Object(can_provide_vars: false);
        command = comment;
    }
    
    public override void construct_command()
    {
    }
    
    public override void parse()
    {
    }
}

public abstract class Euclide.CMake.Function : Euclide.CMake.Command
{
    public Gee.ArrayList<string> args = new Gee.ArrayList<string>();
    public string function_name;
    public string unparsed_args;
    
    construct
    { 
    	can_provide_vars = true;
    }
    
    void check_vars(string text)
    {
        if(text.contains("${"))
        {
            string after_variable_start = text.slice(text.index_of("${") + 2, text.length);
            string variable_name = after_variable_start.slice(0, after_variable_start.index_of("}"));
            vars_needed[variable_name] = "";
            check_vars(after_variable_start.slice(after_variable_start.index_of("}") + 1, after_variable_start.length));
        }
    }

    public override void parse()
    {
        function_name = command.split("(")[0];
        unparsed_args = command.replace(function_name + "(", "");
        function_name = function_name.replace(" ", "");
        unparsed_args = unparsed_args.split(")")[0];
        args.clear();
        foreach(var line in unparsed_args.split("\n"))
        {
            if(line != "")
            {
                /* Don't check after a # */
                string line_content = line.split("#")[0];
                foreach(var arg in line_content.split(" "))
                {
                    if(arg != "")
                    {
                        args.add(arg);
                        check_vars(arg);
                    }
                }
            }
        }
    }
    
    public static Function? create_function(string command)
    {
        string first_part = command.split("(")[0];
        first_part = first_part.replace(" ", "").replace("\t", "");
        switch(first_part)
        {
        case "set":
            return new VariableFunction(command);
        }
        return null;
    }
}
/**
 * A SimpleFunction, it is never really used in the code, it is just here
 * to have a way to create a new #Function
 **/
public class Euclide.CMake.SimpleFunction : Euclide.CMake.Function
{
    public SimpleFunction(string command)
    {
        this.command = command;
        parse();
    }
    
    public override void parse()
    {
    	base.parse();
    }
    
    public override void construct_command()
    {
        command = function_name + "(";
        foreach(var arg in args)
        {
            command += arg + " ";
        }
        command += ")";
    }
}

/**
 * A TargetFunction is a command which is directly related to a target, it can be
 * add_executable, add_library, target_link_libraries…
 **/
public class Euclide.CMake.TargetFunction : Euclide.CMake.Function
{
    public string target_name { get; private set; } /* the name of the target, in 
                                * add_library(euclidecore…), it is euclidecore */
    public string function_value { get; private set; }
    public TargetFunction(string command)
    {
        this.command = command;
        parse();
    }
    
    public override void resolve_vars(Gee.ArrayList<Command> commands)
    {
        base.resolve_vars(commands);
        
        foreach(var entry in vars_needed.entries)
        {
            target_name = target_name.replace("${" + entry.key + "}", entry.value);
            function_value = function_value.replace("${" + entry.key + "}", entry.value);
        }
    }
    
    public override void parse()
    {
    	base.parse();
        target_name = args[0];
        int j = 0;
        function_value = "";
        foreach(var arg in args)
        {
            if(j > 0)
            {
                if(j > 1)
                    function_value += " ";
                function_value += arg;
            }
            j++;
        }
    }
    
    public override void construct_command()
    {
        command = function_name + "(";
        foreach(var arg in args)
        {
            command += arg + " ";
        }
        command += ")";
    }
}

public class Euclide.CMake.TargetLinkLibraries : Euclide.CMake.TargetFunction
{
    public TargetLinkLibraries(string command)
    {
        base(command);
    }
}

/**
 * A VariableFunction is a command which defines a variable. It can be
 * something like set(VAR1 value) or pkg_check_modules(DEPS REQUIRED gtk+-3.0)
 * vala_precompile…
 * 
 * These commands can be used in some other command, like #TargetFunction, 
 * to resolve their names or their dependencies. e.g. for
 * target_link_libraries(euclide ${DEPS_LIBRARIES}) we'll need to search in
 * all VariableFunction of the current CMakeLists.txt to resolve the
 * dependencies names.
 **/
public class Euclide.CMake.VariableFunction : Euclide.CMake.Function
{
    public string variable_name { get; private set; } /* we determine it in parse_function */
    public string variable_value { get; private set; } /* idem */

    /* This list is built in parse_function, don't change it content by hand, it
     * won't be saved anyway, you must modify args.
     */
    public Gee.ArrayList<string> variable_args = new Gee.ArrayList<string>();
    public VariableFunction(string command)
    {
        this.command = command;
        parse();
    }
    
    public override void parse()
    {
    	base.parse();
        variable_name = args[0];
        variable_value = "";
        variable_args.clear();
        foreach(var arg in args)
        {
            if(arg != args[0])
            {
                if(variable_value != "")
                {
                    variable_value += " ";
                }
                variable_args.add(arg);
                variable_value += arg;
            }
        }
        save_defined_vars();
    }
    
    public override void resolve_vars(Gee.ArrayList<Command> commands)
    {
        base.resolve_vars(commands);
        vars_defined.unset(variable_name);
        
        foreach(var entry in vars_needed.entries)
        {
            variable_name = variable_name.replace("${" + entry.key + "}", entry.value);
            variable_value = variable_value.replace("${" + entry.key + "}", entry.value);
        }
        save_defined_vars();
    }
    
    public virtual void save_defined_vars()
    {
        if(function_name == "set")
            vars_defined[variable_name] = variable_value;
    }
    
    public override void construct_command()
    {
        if(function_name == "set" || function_name == "pkg_check_modules")
        {
            command = function_name + "(";
            foreach(var arg in args)
            {
                command += arg + " ";
            }
            command += ")";
        }
    }
}

/**
 * pkg_check_modules
 **/
public class Euclide.CMake.PkgCheckModules : Euclide.CMake.VariableFunction
{
    public List<string> packages;
    
    public PkgCheckModules(string command)
    {
        base(command);
    }
    
    public override void parse()
    {
    	base.parse();
    }
    
    public override void resolve_vars(Gee.ArrayList<Command> commands)
    {
        base.resolve_vars(commands);
    	packages = new List<string>();
        foreach(var arg in variable_value.split(" "))
        {
            if(arg != "REQUIRED")
            {
                packages.append(arg);
            }
        }
    }

    public override void save_defined_vars()
    {
        vars_defined[variable_name + "_LIBRARIES"] = variable_value;
        vars_defined[variable_name + "_CFLAGS"] = variable_value;
        vars_defined[variable_name + "_LIBRARY_DIRS"] = variable_value;
    }
    
    public override void construct_command()
    {
        command = "pkg_check_modules(%s REQUIRED".printf(variable_name);
        foreach(var pack in packages)
            command += " %s\n".printf(pack);
        command += ")";
    }
    
    public bool remove_package(string file)
    {
        for(int i = 0; i < packages.length(); i++)
        {
            if(packages.nth_data(i) == file)
            {
                packages.remove(packages.nth_data(i));
                return true;
            }
        }
        return false;
    }
}


/**
 * vala_precompile
 **/
public class Euclide.CMake.ValaPrecompile : Euclide.CMake.VariableFunction
{
    public List<string> packages = new List<string>();
    public List<string> files = new List<string>();
    public List<string> options = new List<string>();
    
    public ValaPrecompile(string command)
    {
        base(command);
    }
    
    public override void parse()
    {
    	base.parse();
    	
        int state = 3; /* var_name, PACKAGES: 1, OPTIONS: 2, FILES: 3; */
        foreach(var arg in variable_args)
        {
            if(arg == "PACKAGES")
            {
                state = 1;
            }
            else if(arg == "OPTIONS")
            {
                state = 2;
            }
            else if(state == 1)
            {
                packages.append(arg);
            }
            else if(state == 2)
            {
                options.append(arg);
            }
            else if(state == 3)
            {
                files.append(arg);
            }
        }
    }
    
    public override void construct_command()
    {
        command = "vala_precompile(%s\n".printf(variable_name);
        foreach(var file in files)
            command += "    %s\n".printf(file);
        command += "PACKAGES\n";
        foreach(var file in packages)
            command += "    %s\n".printf(file);
        command += "OPTIONS\n";
        foreach(var file in options)
            command += "    %s\n".printf(file);
        command += ")";
    }
    
    public bool remove_file(string file)
    {
        for(int i = 0; i < files.length(); i++)
        {
            if(files.nth_data(i) == file)
            {
                files.remove(files.nth_data(i));
                construct_command();
                return true;
            }
        }
        return false;
    }
    
    public bool add_package(string file)
    {
        foreach(var pack in packages)
        {
            if(pack == file)
            {
                return false;
            }
        }
        packages.append(file);
        construct_command();
        return true;
    }

    public bool remove_package(string file)
    {
        for(int i = 0; i < packages.length(); i++)
        {
            if(packages.nth_data(i) == file)
            {
                packages.remove(packages.nth_data(i));
                construct_command();
                return true;
            }
        }
        return false;
    }
    
    public override void save_defined_vars()
    {
        vars_defined[variable_name] = "ValaPrecompile";
    }
}


public class Euclide.CMake.InstallFunction : Euclide.CMake.Function
{
    public string target;
    bool runtime = false;
    public string destination;
    public InstallFunction(string command)
    {
        this.command = command;
        parse();
    }
    
    public override void parse()
    {
    	base.parse();
        assert(args.size > 2);
        if(args[0] == "TARGET")
        {
            target = args[1];
            if(args[2] == "RUNTIME")
            {
                runtime = true;
                destination = args[4];
            }
            else if(args[2] == "DESTINATION")
            {
                destination = args[3];
            }
        }
    }
    
    public override void construct_command()
    {
        command = function_name + "(";
        foreach(var arg in args)
        {
            command += arg + " ";
        }
        command += ")";
    }
}

/**
 * An assignment command, e.g. BUILD_TYPE=Debug
 * Only used for CMakeCache.
 **/
public class Euclide.CMake.Assignment : Euclide.CMake.Command
{
    /**
     * The key, in BUILD_TYPE=Debug, it is BUILD_TYPE
     **/
    public string key;
    string _assignment_value;

    public string assignment_value
    {
        get { return _assignment_value; }
        set
        {
            _assignment_value = value;
            construct_command();
        }
    }
    
    public Assignment(string command)
    {
        this.command = command;
        parse();
    }
    
    public override void construct_command()
    {
        command = key + "=" + assignment_value + "\n";
    }
    
    public override void parse()
    {
        key = command.split("=")[0];
        assignment_value = command.split("=")[1].replace("\n", "");
    }
}

public class Euclide.CMake.CommandParser : GLib.Object
{
    public string command;
    public string args;
    public string name;
    public Gee.ArrayList<string> split_args;
    public CommandType command_type = CommandType.NOTHING;
    public CommandParser(string command_)
    {
        command = command_;
        split_args = new Gee.ArrayList<string>();
        check_type();
    }
    
    void check_type()
    {
        for(int i = 0; i < command.length; i++)
        {
            if(command[i] == ' ')
            {
            }
            else if(command[i] == '#' || (i > 0 & command[i-1] == '/' & command[i] == '/'))
            {
                command_type = CommandType.COMMENT;
                break;
            }
            else if(command[i] == '=')
            {
                command_type = CommandType.ASSIGNMENT;
                break;
            }
            else if(command[i] == '(')
            {
                command_type = CommandType.FUNCTION;
                break;
            }
        }
        compute_function();
    }
    
    void compute_function()
    {
        switch(command_type)
        {
        case CommandType.ASSIGNMENT:
            name = command.split("=")[0].replace(" ", "").replace("\n", "");
            args = command.split("=")[1].replace("\n", "");
            break;
        case CommandType.FUNCTION:
            name = "";
            int j = 0;
            for(int i = 0; i < command.length; i++)
            {
                if(command[i] == '(')
                {
                    j = i;
                    break;
                }
                else if(command[i] != ' ')
                {
                    name += command[i].to_string();
                }
            }
            
            args = "";
            
            for(int i = j + 1; i < command.length; i++)
            {
                if(command[i] == ')')
                {
                    j = i;
                    break;
                }
                else
                {
                    args += command[i].to_string();
                }
            }
            
            foreach(var line in args.split("\n"))
            {
                foreach(var arg in line.split(" "))
                {
                    if(arg != "")
                    {
                        split_args.add(arg);
                    }
                }
            }
            break;
        }
        
    }
    
    public Command create_command()
    {
        switch(command_type)
        {
            case CommandType.COMMENT:
                return new Comment(command);
            case CommandType.FUNCTION:
                switch(name)
                {
                case "install":
                    return new InstallFunction(command);
                case "vala_precompile":
                    return new ValaPrecompile(command);
                case "pkg_check_modules":
                    return new PkgCheckModules(command);
                case "SET":
                case "set":
                    return new VariableFunction(command);
                case "target_link_libraries":
                    return new TargetLinkLibraries(command);
                case "add_library":
                case "add_executable":
                    return new TargetFunction(command);
                default:
                    return new SimpleFunction(command);
                }
            case CommandType.ASSIGNMENT:
                return new Assignment(command);
        }
        return new Comment(command);
    }
}
   
