/*
 * Copyright (c) 2011 Lucas Baudin <xapantu@gmail.com>
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; see the file COPYING.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

public abstract class Euclide.CMake.BaseParser : GLib.Object
{
    public Gee.ArrayList<Command> commands = new Gee.ArrayList<Command>();
    public string filename;
    public string content;
    public Gee.ArrayList<Parser> children = new Gee.ArrayList<Parser>();
    
    public void load_content()
    {
        content = "";
        try
        {
            FileUtils.get_contents(filename, out content);
        }
        catch(FileError e)
        {
            critical("Couldn't read CMakeLists.txt: %s", e.message);
        }
    }
    
    public void parse_file()
    {
        parse_string(commands, content);
    }
    
    public static void parse_string(Gee.ArrayList<Command> commands, string text)
    {
        string current_command = null;
        string next_stop = null;
        bool reverse = false;
        bool is_function = false;
        for(int i = 0; i < text.length; i++)
        {
            if(text[i] == '=' && !is_function)
            {
                next_stop = "\n";
            }
            if(text[i] == '(') /* to avoid problem with this kind of functions : vala_precompile( --vapidir=\n) */
            {
                is_function = true;
            }
            if(current_command == null)
            {
                reverse = false;
                next_stop = null;
                if(text[i].to_string() != " " && text[i].to_string() !=  "\n")
                {
                    current_command = text[i].to_string();
                    switch(current_command)
                    {
                    case "#":
                        next_stop = "\n";
                        break;
                    case "\n":
                    case " ":
                        reverse = true;
                        next_stop = " ";
                        break;
                    case "/":
                        next_stop = ")";
                        if(i + 1 < text.length & text[i + 1] == '/')
                        {
                            next_stop = "\n";
                        }
                        break;
                    default:
                        next_stop = ")";
                        break;
                    }
                }
            }
            else
            {
                current_command += text[i].to_string();
            }
            if( (! reverse && text[i].to_string() == next_stop) || (reverse && text[i].to_string() != next_stop))
            {
                var parser = new CommandParser(current_command);
                CMake.Command command = parser.create_command();
                if(command is CMake.TargetFunction)
                {
                    ((CMake.TargetFunction)command).resolve_vars(commands);
                }
                command.resolve_vars(commands);
                commands.add(command);
                current_command = null;
                is_function = false;
            }
        }
    }

    public static bool check_extension(string filename)
    {
        string extension = filename.split(".")[filename.split(".").length - 1];
        return extension == "c" || extension == "h" || extension == "vala" || extension == "cpp" || extension == "hpp";
    }
    
    public string get_current_dir()
    {
        return (GLib.File.new_for_path(filename)).get_parent().get_path();
    }
    
    public static VariableFunction? get_variable_value(Gee.ArrayList<Command> commands, string key)
    {
        foreach(var command in commands)
        {
            if(command is VariableFunction)
            {
                VariableFunction variable = (VariableFunction)command;
                if(variable.variable_name == key)
                {
                    return variable;
                }
            }
        }
        return null;
    }
}

public class Euclide.CMake.Parser : Euclide.CMake.BaseParser
{
    public string arguments {set; get;}
    
    public Project project;
    public Cache cache;
    FileMonitor monitor;
    
    public Parser (string filename_, Project? proj = null) {
        try {
            monitor = (GLib.File.new_for_path (filename_)).monitor (FileMonitorFlags.NONE);
            debug ("monitor set on %s\n", filename_);
            monitor.changed.connect ((f, o, e) => { print ("CHANGED: %s\n", e.to_string ()); });
            filename = filename_;
            load_content ();
            parse_file ();
            construct_project (proj);
        } catch (Error e) {
            warning (e.message);
        }
    }

    /**
     * Save the content of the parser in it CMakeLists.txt. All functions modifications
     * will be written.
     **/
    public void save_content()
    {
        string content = "";
        foreach(var command in commands)
        {
            content += command.command + "\n";
        }
        try
        {
            FileUtils.set_contents(filename, content);
        }
        catch(FileError e)
        {
            critical("Couldn't write the CMakeLists.txt: %s", e.message);
        }
        foreach(var child in children)
        {
            child.save_content();
        }
        
    }
    
    /**
     * This function handle all the TargetFunction found in this CMake file, in
     * order to create a project with the good targets, the dependencies (internal
     * and external).
     *
     * @param function a valid TargetFunction
     * @param project a Project
     * @param file_list a list of the files which must be included in this binary,
     * e.g. the vala files which can't be put in add_executable or add_library.
     * @param vala_precompile a VariableFunction, it can be null or contains a
     * vala_precompile function.
     **/ 
    void handle_targets(TargetFunction function, Project project,
                        Gee.ArrayList<File>? file_list = null) {
        string binary_name = function.target_name;
        string binary_path = project.project_path + "/build" + get_current_dir().replace(project.project_path, "") + "/" + binary_name;

        switch(function.function_name) {
        case "add_library":
        case "add_executable":
            var bin  = new Binary(binary_path, binary_name);
            
            /* Transfer vala files */
            bin.files = new Gee.ArrayList<File>();
            bin.files.add_all(file_list);
            file_list.clear();

            /* Save the differents functions related to it */
            
            /* Search for vala_precompile */
            foreach(Command cmd in function.vars_needed_cmd.values)
            {
                if(cmd is ValaPrecompile)
                {
                    foreach(var arg in (cmd as ValaPrecompile).files)
                    {
                        bin.files.add(new File(arg, get_current_dir() + "/" + arg));
                    }
                    bin.related_functions.add(cmd);
                }
            }
            bin.related_functions.add(function);

            /* Save the source dir */
            bin.source_dir = get_current_dir();

            /* Add C files */
            foreach(var arg in function.args)
            {
                if(check_extension(arg))
                {
                    bin.files.add(new File(arg, get_current_dir() + "/" + arg));
                }
            }

            if(function.function_name == "add_library")
                project.lib.add(bin);
            else /* if it is not a library, it must be an executable */
                project.bin.add(bin);

            break;
        case "target_link_libraries":
            Binary? binary = project.get_target(function.target_name);
            if(binary == null) {
                critical("Couldn't find binary name: %s", function.target_name);
            }
            else {
                binary.related_functions.add(function);
            }
            break;
        }
    }
    
    static void create_cmake_file(string file_in, string file_out,
                                  string replace_with = "Project",
                                  string keyword = "$project") {
        string content = "";
        try {
            FileUtils.get_contents(file_in, out content);
            content = content.replace(keyword, replace_with);
            content = content.replace(keyword.up(), replace_with.up());
            FileUtils.set_contents(file_out, content);
        }
        catch (Error e) {
            critical("Hum, check these files: %s %s", file_in, file_out);
        }
    }
    
    public bool add_target (string name, bool library = false) {
        foreach (var bin in project.bin) {
            if(bin.name == name)
                return false;
        }
        foreach (var bin in project.lib) {
            if(bin.name == name)
                return false;
        }
        DirUtils.create (project.project_path + "/" + name, 511);
        string prefix = library ? "library_" : "executable_";
        create_cmake_file ("%s/tpl/%sCMakeLists.txt".printf (Build.RESSOURCES_DIR, prefix),
                          project.project_path + "/" + name + "/CMakeLists.txt",
                          name);

        create_cmake_file ("%s/tpl/%smain.vala".printf(Build.RESSOURCES_DIR, prefix),
                          project.project_path + "/" + name + "/main.vala",
                          name);

        /* Add our subdirectory */
        commands.add (new SimpleFunction("add_subdirectory (" + name + ")"));
        save_content ();

        commands.clear ();
        children.clear ();
        load_content ();
        parse_file ();
        construct_project (null);
        return true;
    }

    public void construct_project (Project? proj = null) {
        project = proj ?? new Project();
        bool root = false;
        if(proj == null)
        {
            root = true;
            project.project_path = get_current_dir();
        }
        var file_list = new Gee.ArrayList<File>();
        foreach(var command in commands)
        {
            if(command is Function)
            {
                Function function = (Function)command;
                if(function is TargetFunction)
                {
                    handle_targets(function as TargetFunction, project, file_list);
                }
                else if(function is ValaPrecompile)
                {
                    project.is_vala = true;
                }
                else
                {
                    switch(function.function_name)
                    {
                    case "project":
                        project.name = function.unparsed_args;
                        break;
                    case "cmake_minimum_required":
                        project.cmake_version = function.unparsed_args;
                        break;
                    case "cmake_policy":
                        project.cmake_policy = function.unparsed_args;
                        break;
                    /*case "list":
                        if(function.args.size == 3)
                        {
                            if(function.args[0] == "APPEND" &&
                               function.args[1] == "CMAKE_MODULE_PATH" &&
                               function.args[2] == "${CMAKE_SOURCE_DIR}/cmake/vala")
                            {
                                project.is_vala = true;
                            }
                        }
                        break;*/
                    case "add_subdirectory":
                        string current_dir = get_current_dir();
                        var child = new Parser(current_dir + "/" + function.unparsed_args.replace("\"", "") + "/CMakeLists.txt", project);
                        children.add(child);
                        break;
                    }
                }
            }
        }
        
        if(root)
        {
            string cache_file = get_current_dir() + "/build/CMakeCache.txt";
            if(FileUtils.test(cache_file, FileTest.EXISTS))
            {
                cache = new Cache(cache_file);
            }
            else
            {
                //debug("No project configuration.");
            }
        }
    }
    
}
