/*
 * Copyright (c) 2011 Lucas Baudin <xapantu@gmail.com>
 *
 * This is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; see the file COPYING.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

public class Euclide.BuildActions.Build : GLib.Object
{
    public signal void build_finished(int state);
    RawTerminal raw_terminal = null;
    Terminal terminal = null;
    Gtk.ProgressBar progress = null;
    Threads.SimpleCommand simple_command;
    Threads.SimpleCommand cmake_command = null;
    string project_path;

    public Build(CMake.Parser parser, RawTerminal? raw_terminal = null, Terminal? terminal = null, Gtk.ProgressBar? progress = null)
    {
        if(raw_terminal != null)
        {
            raw_terminal.clear();
            this.raw_terminal = raw_terminal;
        }
        if(terminal != null)
        {
            terminal.clear();
            this.terminal = terminal;
        }
        if(progress != null)
        {
            progress.set_fraction(0.0);
            progress.show();
            this.progress = progress;
        }
        project_path = parser.get_current_dir();
        bool is_configured = FileUtils.test(project_path + "/build/CMakeCache.txt", FileTest.EXISTS);
        if(!is_configured)
        {
            try {
                Process.spawn_sync(project_path, {"mkdir", "build", "-p"}, null, SpawnFlags.SEARCH_PATH, null);
            } catch (SpawnError e) {
                warning (e.message);
            }
            cmake_command = new Threads.SimpleCommand(project_path + "/build/", "/usr/bin/cmake ..");
            cmake_command.done.connect( () => { simple_command.run(); });
        }
        simple_command = new Threads.SimpleCommand(project_path + "/build/", "/usr/bin/make");
        simple_command.output_changed.connect(terminal_output);
        simple_command.done.connect((exit) => { this.progress.hide(); build_finished(exit); });
    }
    
    public void run()
    {
        if(cmake_command != null)
        {
            cmake_command.run();
        }
        else
        {
            simple_command.run();
        }
    }
    
    void terminal_output(string text)
    {
        raw_terminal.add_text(text);
        Gee.ArrayList<Output.ParserResult?> results = Output.parse(text);
        foreach(var result in results)
        {
            if(result.type == Output.ResultType.PROGRESS)
            {
                progress.set_fraction(((double)result.value)/100.0);
            }
            else if(result.type == Output.ResultType.WARNING || result.type == Output.ResultType.ERROR)
            {
				switch(result.type){
				case Output.ResultType.ERROR:
					terminal.add_error(result.text, result.file, result.line, project_path);
					break;
				case Output.ResultType.WARNING:
					terminal.add_warning(result.text, result.file, result.line, project_path);
					break;
				default:
					terminal.add_question(result.text, result.file, result.line, project_path);
					break;
				}
            }
        }
    }
}

public class Euclide.BuildActions.Check : GLib.Object
{
    public signal void build_finished(int state);
    RawTerminal raw_terminal = null;
    Threads.SimpleCommand simple_command;
    string project_path;

    public Check(CMake.Parser parser, RawTerminal? raw_terminal = null)
    {
        if(raw_terminal != null)
        {
            raw_terminal.clear();
            this.raw_terminal = raw_terminal;
        }
        project_path = parser.get_current_dir();
        simple_command = new Threads.SimpleCommand(project_path + "/build/", "/usr/bin/make check");
        simple_command.output_changed.connect(terminal_output);
        simple_command.done.connect((exit) => { build_finished(exit); });
    }
    
    public void run()
    {
        simple_command.run();
    }
    
    void terminal_output(string text)
    {
        raw_terminal.add_text(text);
    }
}
