/*
 * Copyright (c) 2011 Lucas Baudin <xapantu@gmail.com>
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; see the file COPYING.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

using Gtk;

class Euclide.CloseButton : Button
{
    public string item;
    public CloseButton(string item)
    {
        this.item = item;
        var hbox = new HBox(false, 0);
        hbox.pack_end(new Gtk.Image.from_stock(Gtk.Stock.CLOSE, Gtk.IconSize.MENU), false, false);
        hbox.pack_start(new Label(item));
        add(hbox);
        
        margin_right = margin_left = margin_top = 5;
    }
}

class Euclide.DepsBox : Box
{
    public signal void item_removed(string item);
    public signal void item_added(string item);
    Button add_button;
    Entry entry;
    Label no_deps;
    Granite.Widgets.FlowBox flowbox;
    List<Gtk.Widget> deps_children;
    public DepsBox()
    {
        deps_children = new List<Gtk.Widget>();
        flowbox = new Granite.Widgets.FlowBox ();
        spacing = 2;
        add_button = new Button.with_label(_("Add"));
        pack_end(add_button, false, false);
        add_button.clicked.connect(() => { new_item(); });
        entry = new Entry();
        pack_end(entry);
        entry.hide();
        entry.no_show_all = true;
        no_deps = new Label(_("(no dependencies detected)"));
        pack_start(no_deps);
        pack_start(flowbox);
    }
    
    public void add_item(string item)
    {
        if(no_deps != null)
        {
            no_deps.destroy();
        }
        if(item == "")
            return;
        var button = new CloseButton(item);
        flowbox.add(button);
        deps_children.append(button);
        button.clicked.connect(() => { remove_item(item); });
        show_all();
    }
    
    void new_item()
    {
        add_button.hide();
        entry.activate.connect( () => {entry.hide();});
        entry.focus_out_event.connect( () => {add_button.show(); item_added(entry.text); add_item(entry.text); entry.hide(); entry.text = ""; return false; });
        entry.show();
        entry.grab_focus();
    }

    public void remove_item(string item)
    {
        foreach(var child in deps_children)
        {
            if(child is CloseButton)
            if((child as CloseButton).item == item)
            {
                child.destroy();
                item_removed(item);
            }
        }
    }
}

public class Euclide.ProjectSettingsDialog : HBox
{
    CMake.Parser parser;
    bool build_other = false;
    Box vbox;
    VBox vbox_deps;
    Grid grid;
    int row = 0;
    
    public ProjectSettingsDialog(CMake.Parser parser)
    {
        this.parser = parser;
        vbox = new Gtk.VBox(false, 2);
        //pack_start(vbox, true, true);
        
        grid = new Gtk.Grid();
        grid.row_spacing = 5;
        grid.column_spacing = 5;
        grid.margin_left = 12;
        grid.margin_right = 12;
        grid.margin_top = 12;
        grid.margin_bottom = 12;     
        pack_start(grid, true, true);
        
        add_build_target();
        add_deps();
    }
    
    void add_build_target()
    {
        string build_type = parser.cache.get_value(CMake.Cache.BUILD_TYPE);

        var label_personal = new Label("");
        label_personal.use_markup = true;
        label_personal.set_label("<b>" + _("Personal Settings:") +"</b>");
        add_section (grid, label_personal, ref row);
        
        var check = new Granite.Widgets.ModeButton();
        check.append(new Label(_("Release")));
        check.append(new Label(_("Debug")));
        check.hexpand = false;
        switch(build_type)
        {
        case "":
        case "Release":
            check.selected = 0;
            break;
        case "Debug":
            check.selected = 1;
            break;
        default:
            check.append(new Label(_("Other")));
            check.selected = 2;
            build_other = true;
            break;
        }
        check.mode_changed.connect( (w) => {
            if(build_other)
                check.remove(2);
            parser.cache.set_value(CMake.Cache.BUILD_TYPE, (w as Label).label);
        });
        var label = new Label("Project Build Type:");
        add_option(grid, label, check, ref row);
        
        var label_make_ctest = new Label ("Use \"make check\" instead of \"ctest\":");
        var check_ctest = new Switch ();
        euclide_settings.bind ("make-ctest", check_ctest, "active", SettingsBindFlags.DEFAULT);
        add_option(grid, label_make_ctest, check_ctest, ref row);
        
        row++;

    }
    
    void add_deps()
    {
        vbox_deps = new Gtk.VBox(false, 2);
        vbox.pack_start(vbox_deps);
        var label_personal = new Label("");
        label_personal.use_markup = true;
        label_personal.set_label("<b>" + _("Project Settings:") + "</b>");
        add_section (grid, label_personal, ref row); 

        foreach(var bin in parser.project.bin)
        {
            add_binary(bin);
        }
        foreach(var bin in parser.project.lib)
        {
            add_binary(bin);
        }
        row++;
    }
    
    void add_binary(CMake.Binary binary)
    {
        var label = new Label("");
        label.set_alignment(0.0f, 0.5f);
        label.use_markup = true;
        label.set_label(_("Dependencies of %s:").printf("<b>" + binary.name + "</b>"));

        var flow = new DepsBox();
        foreach(var deps in binary.get_dependencies())
        {
            flow.add_item(deps);
        }
        flow.item_removed.connect((s) => {
            binary.remove_dep(s);
            parser.save_content();
            vbox_deps.destroy();
            add_deps();
            show_all();
        });
        flow.item_added.connect((s) => {
            binary.add_dep(s);
            parser.save_content();
            vbox_deps.destroy();
            add_deps();
            show_all();
        });
        add_option(grid, label, flow, ref row);
    }
    
    void add_section (Gtk.Grid grid, Gtk.Label name, ref int row) {
        name.use_markup = true;
        name.set_markup ("<b>%s</b>".printf (name.get_text ()));
        name.halign = Gtk.Align.START;
        grid.attach (name, 0, row, 1, 1);
        row ++;
    }
        
    void add_option (Gtk.Grid grid, Gtk.Widget label, Gtk.Widget switcher, ref int row) {
        label.hexpand = true;
        label.halign = Align.END;
        label.margin_left = 20;
        switcher.halign = Gtk.Align.FILL;
        switcher.hexpand = true;
          
        if (switcher is Switch || switcher is CheckButton
            || switcher is Entry) { /* then we don't want it to be expanded */
            switcher.halign = Gtk.Align.START;
        }
            
        grid.attach (label, 0, row, 1, 1);
        grid.attach_next_to (switcher, label, Gtk.PositionType.RIGHT, 3, 1);
        row ++;
    }
}

public class Euclide.EntryDialog : Dialog
{
    Granite.Widgets.HintedEntry entry_filename;
    public string value_entry { owned get { return entry_filename.text;  }}
    public EntryDialog(Gtk.Window parent)
    {
        add_buttons(Gtk.Stock.OK, Gtk.ResponseType.OK);
        entry_filename = new Granite.Widgets.HintedEntry("Filename.vala");
        ((Gtk.Box)get_content_area()).add(entry_filename);
        show_all();
    }
}
