/*
 * Copyright (c) 2011 Lucas Baudin <xapantu@gmail.com>
 *               2013 Mario Guerriero <mario@elementaryos.org>
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; see the file COPYING.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

namespace Euclide {
    
    /**
     * Class for easily dealing with files.
     */
    internal class File : GLib.Object {

        public GLib.File file;
        public bool lib;
        private GLib.FileInfo info;

        private enum Type {
            VALID_FILE,
            VALID_FOLDER,
            UNKNOWN,
            INVALID
        }

        public File (string path, string? name = null, bool lib = false) {
            file = GLib.File.new_for_path (path);
            if (lib) this.name = name + " (lib)";
            else this.name = name;
            this.lib = lib;
            
            info = new FileInfo ();
            try {
                info = file.query_info (
                    GLib.FileAttribute.STANDARD_CONTENT_TYPE + "," +
                    GLib.FileAttribute.STANDARD_IS_BACKUP + "," +
                    GLib.FileAttribute.STANDARD_IS_HIDDEN + "," +
                    GLib.FileAttribute.STANDARD_DISPLAY_NAME + "," +
                    GLib.FileAttribute.STANDARD_TYPE,
                    FileQueryInfoFlags.NONE);
            } catch (GLib.Error error) {
                info = null;
                warning (error.message);
            }
            
            if (name == null)
                this.name = info.get_display_name ();
        }

        // returns the path the file
        string _path = null;
        public string path {
            get { return _path != null ? _path : _path = file.get_path (); }
        }

        // returns the basename of the file
        public string name;/* {
            get { 
                if (lib)
                    return _name != null ? _name : _name = info.get_display_name () + " (lib)";
                else
                    return _name != null ? _name : _name = info.get_display_name (); 
                }
        }*/

        // returns the icon of the file's content type
        GLib.Icon _icon = null;
        public GLib.Icon icon {
            get {
                if (_icon != null)
                    return _icon;
                //var content_type = info.get_attribute_string (FileAttribute.STANDARD_FAST_CONTENT_TYPE);
                var content_type = info.get_content_type ();
                return _icon = GLib.ContentType.get_icon (content_type);
            }
        }

        // checks if file exists
        public bool exists {
            get { return file.query_exists (); }
        }

        Type _type = Type.UNKNOWN;
        // checks if we're dealing with a non-hidden, non-backup directory
        public bool is_valid_directory {
            get {
                if (_type == Type.VALID_FILE)
                    return false;
                if (_type == Type.VALID_FOLDER)
                    return true;
                if (_type == Type.INVALID)
                    return false;

                if (info.get_file_type () != FileType.DIRECTORY ||
                    info.get_is_hidden () || info.get_is_backup ()) {
                    return false;
                }

                bool has_valid_children = false;

                foreach (var child in children) {
                    if (child.is_valid_textfile) {
                        _type = Type.VALID_FOLDER;
                        return has_valid_children = true;
                    }
                }

                foreach (var child in children) {
                    if (child.is_valid_directory) {
                        has_valid_children = true;
                        _type = Type.VALID_FOLDER;
                    return has_valid_children = true;
                    }
                }

                return false;
            }
        }

        // checks if we're dealing with a textfile
        public bool is_valid_textfile {
            get {
                if (_type == Type.VALID_FILE)
                    return true;
                if (_type == Type.VALID_FOLDER)
                    return false;
                if (_type == Type.INVALID)
                    return false;

                if (info.get_file_type () == FileType.REGULAR) {
                    //var content_type = info.get_attribute_string (FileAttribute.STANDARD_FAST_CONTENT_TYPE);
                    var content_type = info.get_content_type ();
                    if (ContentType.is_a (content_type, "text/*") &&
                        !info.get_is_backup () &&
                        !info.get_is_hidden ()) {
                        _type = Type.VALID_FILE;
                        return true;
                    }
                }

                return false;
            }
        }

        // returns a list of all children of a directory
        GLib.List <File> _children = null;
        public GLib.List <File> children {
            get {
                if (_children != null)
                    return _children;

                var parent = GLib.File.new_for_path (file.get_path ());
                try {
                    var enumerator = parent.enumerate_children (
                        GLib.FileAttribute.STANDARD_NAME,
                        FileQueryInfoFlags.NONE
                    );

                    var file_info = new FileInfo ();
                    while ((file_info = enumerator.next_file ()) != null) {
                        var child = parent.get_child (file_info.get_name ());
                        _children.append (new File (child.get_path ()));
                    }
                } catch (GLib.Error error) {
                    warning (error.message);
                }

                return _children;
            }
        }

        /*public void rename (string name) {
            try {
                file.set_display_name (name);
            } catch (GLib.Error error) {
                warning (error.message);
            }
        }*/

        public void trash () {
            try {
                file.trash ();
            } catch (GLib.Error error) {
                warning (error.message);
            }
        }

        public void reset_cache () {
            name = null;
            _path = null;
            _icon = null;
            _children = null;
            _type = Type.UNKNOWN;
        }

        public static int compare (File a, File b) {
            if (a.is_valid_directory && b.is_valid_textfile)
                return -1;
            if (a.is_valid_textfile && b.is_valid_directory)
                return 1;
            return strcmp (a.path.collate_key_for_filename (),
                           b.path.collate_key_for_filename ());
        }
    }
    
    internal class ProgramList : Gtk.Box {
        public Gtk.ProgressBar progress;
        public FileList list;
        
        public signal void open_file (string path);
        
        public ProgramList () {
            this.orientation = Gtk.Orientation.VERTICAL;
            
            list = new FileList ();
            progress = new Gtk.ProgressBar ();
            progress.visible = false;
            progress.no_show_all = true;
        
            this.pack_start (list, true, true, 0);
            this.pack_start (progress, false, true, 0);
            
            list.open_file.connect ((p) => { this.open_file (p); });
        }
        
        public void open_project (Euclide.CMake.Parser cmake, string? project_path = null) {
            list.open_project (cmake, project_path);
        }
    }
    
    /**
     * SourceList that displays folders and their contents.
     */
    Gee.ArrayList<Euclide.CMake.Parser> parsers;
    internal class FileList : Granite.Widgets.SourceList {

        public signal void open_file (string path);

        public FileList () {
            this.width_request = 180;
        
            parsers = new Gee.ArrayList<Euclide.CMake.Parser> ();
            
            this.item_selected.connect ((item) => {
                open_file ((item as FileItem).path);
            });

            this.set_sort_func ((a, b) => {
                return File.compare ((a as Item).file, (b as Item).file);
            });
        }

        public void open_project (Euclide.CMake.Parser cmake, string? project_path = null) {
            
            if (parsers.contains (cmake))
                return;
            
            parsers.add (cmake);
            
            if (project_path != null)
                this.add_folder (new File (project_path), true);
        }

        private void add_folder (File folder, bool expand) {
            if (is_open (folder)) {
                warning ("Folder '%s' is already open.", folder.path);
                return;
            } else if (!folder.is_valid_directory) {
                warning ("Cannot open invalid directory.");
                return;
            }

            var folder_root = new FolderItem (folder);
            this.root.add (folder_root);

            folder_root.expanded = expand;
        }

        private bool is_open (File folder) {
            foreach (var child in root.children)
                if (folder.path == (child as Item).path)
                    return true;
            return false;
        }
    }

    internal class IconLoader {
        private static IconLoader instance;

        private int icon_size;
        private Gee.HashMap<string, unowned Gdk.Pixbuf> icon_cache;
        private Gtk.StyleContext style;

        private IconLoader () {
            icon_cache = new Gee.HashMap<string, unowned Gdk.Pixbuf> ();
            style = new Gtk.StyleContext ();

            int width, height;
            Gtk.icon_size_lookup (Gtk.IconSize.MENU, out width, out height);
            icon_size = int.max (width, height);
        }

        public static IconLoader get_default () {
            if (instance == null)
                instance = new IconLoader ();
            return instance;
        }

        public Gdk.Pixbuf get_rendered_icon (GLib.Icon icon) {
            string key = icon.to_string ();

            var cached_pixbuf = icon_cache.get (key);

            if (cached_pixbuf != null)
                return cached_pixbuf;

            var factory = Granite.Services.IconFactory.get_default ();
            var new_pixbuf = factory.load_symbolic_icon_from_gicon (style, icon, icon_size);
            icon_cache.set (key, new_pixbuf);

            return new_pixbuf;
        }
    }

    /**
     * Common interface for normal and expandable items.
     */
    internal interface Item : Granite.Widgets.SourceList.Item {
        public abstract File file { get; construct; }
        public abstract string path { get; }
    }

    /**
     * Normal item in the source list, represents a textfile.
     * TODO Remove, Rename
     */
    internal class FileItem : Granite.Widgets.SourceList.Item, Item {

        Gtk.Menu menu;
        Gtk.MenuItem item_delete;

        public File file { get; construct; }
        public string path { get { return file.path; } }

        public FileItem (File file) requires (file.is_valid_textfile) {
            Object (file: file);

            this.selectable = true;
            //this.editable = true;
            this.name = file.name;
            this.icon = IconLoader.get_default ().get_rendered_icon (file.icon);
        }

        /*public void rename (string new_name) {
            file.rename (new_name);
        }*/

        public override Gtk.Menu? get_context_menu () {
            menu = new Gtk.Menu ();
            item_delete = new Gtk.MenuItem.with_label (_("Delete"));
            menu.append (item_delete);
            item_delete.activate.connect (() => { 
                file.trash ();
                // Remove the file from the targets
                foreach (var cmake in parsers) {
                    
                    foreach (var bin in cmake.project.bin)
                        foreach (var f in bin.files)
                            if (f.absolute == file.file.get_path ()) {
                                string file = (f.absolute).replace (bin.source_dir, "");
                                bin.remove_file (file[1:file.length]);
                            }
                    
                    foreach (var lib in cmake.project.lib)
                        foreach (var f in lib.files)
                            if (f.absolute == file.file.get_path ()) {
                                string file = (f.absolute).replace (lib.source_dir, "");
                                lib.remove_file (file[1:file.length]);
                            }
                    
                    cmake.save_content();
                }
                // TODO: implement a new function in Scratch's API to close from uri/path
                //plugins.close_document (doc);
            });
            menu.show_all ();
            return menu;
        }
    }

    /**
     * Expandable item in the source list, represents a folder.
     * Monitored for changes inside the directory.
     * TODO remove, rename, create new file
     */
    internal class FolderItem : Granite.Widgets.SourceList.ExpandableItem, Item {

        Gtk.Menu menu;
        Gtk.MenuItem item_create;

        private GLib.FileMonitor monitor;
        private bool children_loaded = false;

        public File file { get; construct; }
        public string path { get { return file.path; } }

        public FolderItem (File file) requires (file.is_valid_directory) {
            Object (file: file);

            this.editable = false;
            this.selectable = false;
            this.name = file.name;
            this.icon = IconLoader.get_default ().get_rendered_icon (file.icon);

            this.add (new Granite.Widgets.SourceList.Item ("")); // dummy
            this.toggled.connect (() => {
                if (this.expanded && this.n_children <= 1) {
                    this.clear ();
                    this.add_children ();
                    children_loaded = true;
                }
            });

            try {
                monitor = file.file.monitor_directory (GLib.FileMonitorFlags.NONE);
                monitor.changed.connect ((s,d,e) => { on_changed (s,d,e); });
            } catch (GLib.Error e) {
                warning (e.message);
            }
        }

        public override Gtk.Menu? get_context_menu () {
            menu = new Gtk.Menu ();
            item_create = new Gtk.MenuItem.with_label (_("Add new File"));
            menu.append (item_create);
            item_create.activate.connect (() => {
                var dialog = new Gtk.Dialog ();
                dialog.title = _("Add new file");
                
                var entry = new Gtk.Entry ();
                entry.activate.connect (() => { dialog.response (0); });
                
                var target_box = new Gtk.Box (Gtk.Orientation.HORIZONTAL, 5);
                
                var combo_target = new Gtk.ComboBoxText ();
                combo_target.append_text (_("Plain File"));
                foreach (var cmake in parsers) {
                    foreach (var bin in cmake.project.bin) {
                        combo_target.append_text (bin.name);
                    }
                    foreach (var lib in cmake.project.lib) {
                        combo_target.append_text (lib.name + " (lib)");
                    }
                }
                combo_target.active = 1;
                
                target_box.pack_start (new Gtk.Label (_("Target:")), false, true, 5);
                target_box.pack_start (combo_target, true, true, 5);
                
                ((Gtk.Box)dialog.get_content_area ()).pack_start (entry, true, true, 5);
                ((Gtk.Box)dialog.get_content_area ()).pack_start (target_box, true, true, 5);
                dialog.add_button (_("Add"), 0);
                dialog.show_all ();
                
                dialog.response.connect (() => {
                    try {
                        var new_file = GLib.File.new_for_path (file.path + "/" + entry.text);
		                new_file.create (FileCreateFlags.NONE);
		                // Add to the proper target
		                if (combo_target.active != 0)
		                    foreach (var cmake in parsers) {
                                string target = combo_target.get_active_text ();
                                foreach (var bin in cmake.project.bin)
                                    if (bin.name == target) {
                                        string file = (file.path + "/" + entry.text).replace (bin.source_dir, "");
                                        bin.add_file (file[1:file.length]);
                                    }
                                foreach (var lib in cmake.project.lib)
                                    if ((lib.name + " (lib)") == target) {
                                        string file = (file.path + "/" + entry.text).replace (lib.source_dir, "");
                                        lib.add_file (file[1:file.length]);
                                    }
                                cmake.save_content();
                            }
	                } catch (Error e) {
		                warning ("Error: %s\n", e.message);
	                }
                    dialog.destroy ();
                });
                
                dialog.run ();
                
            });
            menu.show_all ();
            return menu;
        }

        internal void add_children () {
            foreach (var child in file.children) {
                if (child.is_valid_directory) {
                    var item = new FolderItem (child);
                    add (item);
                } else if (child.is_valid_textfile) {
                    var item = new FileItem (child);
                    add (item);
                    //item.edited.connect (item.rename);
                }
            }
        }

        private void on_changed (GLib.File source, GLib.File? dest, GLib.FileMonitorEvent event) {

            if (!children_loaded) {
                this.file.reset_cache ();
                return;
            }

            switch (event) {

                case GLib.FileMonitorEvent.DELETED:
                    foreach (var item in children)
                        if ((item as Item).path == source.get_path ())
                            remove (item);
                    break;

                case GLib.FileMonitorEvent.CREATED:
                    if (source.query_exists () == false) {
                        return;
                    }
                    var file = new File (source.get_path ());
                    var exists = false;
                    foreach (var item in children)
                        if ((item as Item).path == file.path)
                            exists = true;

                    if (!exists) {
                        if (file.is_valid_textfile)
                            this.add (new FileItem (file));
                        else if (file.is_valid_directory)
                            this.add (new FolderItem (file));
                        // this.file.reset_cache ();
                    }
                    break;
            }
        }
    }
}
