/*
 * Copyright (c) 2011 Lucas Baudin <xapantu@gmail.com>
 *
 * This is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; see the file COPYING.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

using Euclide;

namespace Euclide.Tests {

    void add_cmake_tests () {
        Test.add_func ("/euclide/cmake/command", () => {
            var parser = new CMake.CommandParser("#test");
            CMake.Command command = parser.create_command();
            assert(command is CMake.Comment);
            assert(command.command == "#test");

            parser = new CMake.CommandParser(" #test");
            command = parser.create_command();
            assert(command is CMake.Comment);
            assert(command.command == " #test");

            parser = new CMake.CommandParser("//test");
            command = parser.create_command();
            assert(command is CMake.Comment);
            assert(command.command == "//test");

            parser = new CMake.CommandParser(" test()");
            command = parser.create_command();
            assert(command is CMake.Function);
            CMake.Function function = (CMake.Function)command;
            assert(function.command == " test()");
            assert(function.function_name == "test");
            assert(function.unparsed_args == "");
            assert(function.args.size == 0);

            parser = new CMake.CommandParser("test(2.5.2) ");
            command = parser.create_command();
            assert(command is CMake.Function);
            function = (CMake.Function)command;
            assert(function.function_name == "test");
            assert(function.unparsed_args == "2.5.2");
            assert(function.args.size == 1);
            assert(function.args[0] == "2.5.2");

            parser = new CMake.CommandParser("test(2.=5.2) ");
            command = parser.create_command();
            assert(command is CMake.Function);
            function = (CMake.Function)command;
            assert(function.function_name == "test");
            assert(function.unparsed_args == "2.=5.2");
            assert(function.args.size == 1);
            assert(function.args[0] == "2.=5.2");

            parser = new CMake.CommandParser("   test(2.5.2 VERSION) ");
            command = parser.create_command();
            assert(command is CMake.Function);
            function = (CMake.Function)command;
            assert(function.function_name == "test");
            assert(function.unparsed_args == "2.5.2 VERSION");
            assert(function.args.size == 2);
            assert(function.args[0] == "2.5.2");
            assert(function.args[1] == "VERSION");

            parser = new CMake.CommandParser("   ");
            command = parser.create_command();
            assert(command is CMake.Comment);

            parser = new CMake.CommandParser("CMAKE_CACHE=value");
            command = parser.create_command();
            assert(command is CMake.Assignment);
            CMake.Assignment assignment = (CMake.Assignment)command;
            assert(assignment.key == "CMAKE_CACHE");
            assert(assignment.assignment_value == "value");

            parser = new CMake.CommandParser(" install(TARGET euclide RUNTIME DESTINATION bin)");
            command = parser.create_command();
            assert(command is CMake.InstallFunction);
            CMake.InstallFunction install = (CMake.InstallFunction)command;
            assert(install.target == "euclide");
            assert(install.destination == "bin");

            parser = new CMake.CommandParser(" set(VALA_C euclide.c)");
            command = parser.create_command();
            assert(command is CMake.VariableFunction);
            CMake.VariableFunction variable = (CMake.VariableFunction)command;
            assert(variable.variable_name == "VALA_C");
            assert(variable.variable_value == "euclide.c");

            parser = new CMake.CommandParser("SET (VALA_C euclide.c)");
            command = parser.create_command();
            assert(command is CMake.VariableFunction);
        });

        Test.add_func ("/euclide/cmake/parser/string", () => {
            var commands = new Gee.ArrayList<CMake.Command>();
            CMake.BaseParser.parse_string(commands, 
                """#command


                """);
            assert(commands.size == 1);
            assert(commands[0] is CMake.Comment);
            
            commands.clear();
            CMake.BaseParser.parse_string(commands, 
                """#command

                #comment
                """);
            assert(commands.size == 2);
            assert(commands[0] is CMake.Comment);
            assert(commands[1] is CMake.Comment);
            
            commands.clear();
            CMake.BaseParser.parse_string(commands, 
                """# Check http://webdev.elementaryos.org/docs/developer-guide/cmake for documentation

                cmake_minimum_required (VERSION 2.6)
                cmake_policy(VERSION 2.6)

                project(IDE)""");
            assert(commands.size == 4);
            assert(commands[0] is CMake.Comment);
            assert(commands[1] is CMake.Function);
            assert(commands[2] is CMake.Function);
            assert(commands[3] is CMake.Function);
            
            commands.clear();
            CMake.BaseParser.parse_string(commands, 
                """
                # Check http://webdev.elementaryos.org/docs/developer-guide/cmake for documentation

                cmake_minimum_required (VERSION 2.6)
                //cmake_policyVERSION 2.6)
                //cmake_minimum_required VERSION 2.6(
                cmake_minimum_required (VERSION 2.6)

                project(IDE)""");
            assert(commands.size == 6);
            assert(commands[0] is CMake.Comment);
            assert(commands[1] is CMake.Function);
            assert(commands[2] is CMake.Comment);
            assert(commands[3] is CMake.Comment);
            assert(commands[4] is CMake.Function);
            assert(commands[5] is CMake.Function);
        });

        Test.add_func ("/euclide/cmake/parser/equal", () => {
            var commands = new Gee.ArrayList<CMake.Command>();
            CMake.BaseParser.parse_string(commands, 
                """
                vala_precompile( VALA_C
                test.vala
                OPTIONS
                --thread
                --vapidir=
                )


                """);
            assert(commands.size == 1);
            assert(commands[0] is CMake.VariableFunction);
            assert(commands[0].command.contains("vala_precompile"));
        });

        Test.add_func ("/euclide/cmake/parser/variable", () => {
            CMake.Parser parser = new CMake.Parser("../../CMakeLists.txt");
            assert(CMake.BaseParser.get_variable_value(parser.commands, "MAYA_SCHEMAS") != null);
            assert(parser.project.lib[1].name == "euclidecore");
        });

        Test.add_func ("/euclide/cmake/parser/file", () => {
            var cmake = new CMake.Parser("../../CMakeLists.txt");
            assert(cmake.children.size == 3);
            
            assert(cmake.project.name == "Euclide");
            assert(cmake.project.cmake_version == "VERSION 2.6");
            assert(cmake.project.cmake_policy == "VERSION 2.6");
            assert(cmake.project.is_vala == true);
            assert(cmake.project.lib[0].files[0].relative == "main.vala");
            assert(cmake.project.lib[0].name == "euclide");
            assert(cmake.project.lib[0].filename == Environment.get_current_dir() + "/euclide");
        });
        
        Test.add_func ("/euclide/cmake/output_parser", () => {
            Output.ParserResult result = Output.parse("[ 18%] Building C object src/CMakeFiles/ide.dir/main.c.o")[0];
            assert(result.type == Output.ResultType.PROGRESS);
            assert(result.value == 18);
            result = Output.parse(
    """/home/xapantu/Dev/program/src/cmake.vala:150.9-150.53: warning: unhandled error `GLib.FileError'
            FileUtils.get_contents(filename, out content);
            ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^""")[0];
            assert(result.type == Output.ResultType.WARNING);
            assert(result.line == 150);
            assert(result.file == "/home/xapantu/Dev/program/src/cmake.vala");
            result = Output.parse(
    """/home/xapantu/Dev/program/src/cmake.vala:150.9-150.53: error: unhandled error `GLib.FileError'
            FileUtils.get_contents(filename, out content);
            ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^""")[0];
            assert(result.type == Output.ResultType.ERROR);
        });
        
        Test.add_func("/euclide/cmake/cache", () => {
            var cache = new CMake.Cache("../CMakeCache.txt");
            int count = 0;
            string build_type = "";
            foreach(var command in cache.commands)
            {
                if(command is CMake.Assignment)
                {
                    CMake.Assignment assignment = (CMake.Assignment)command;
                    switch(count)
                    {
                    case 0:
                        assert(assignment.key == "CMAKE_AR:FILEPATH" || assignment.key == "BUILD_STATIC:UNINITIALIZED");
                        if(assignment.key == "BUILD_STATIC:UNINITIALIZED") count--;
                        break;
                    case 1:
                        assert(assignment.key == "CMAKE_BUILD_TYPE:STRING");
                        build_type = assignment.assignment_value;
                        break;
                    }
                    count++;
                }
            }
            assert(count > 0);
            cache.set_value(CMake.Cache.BUILD_TYPE, "Debug");
            assert(cache.get_value(CMake.Cache.BUILD_TYPE) == "Debug");
            cache = new CMake.Cache("../CMakeCache.txt");
            assert(cache.get_value(CMake.Cache.BUILD_TYPE) == "Debug");
            cache.set_value(CMake.Cache.BUILD_TYPE, build_type);
            assert(cache.get_value(CMake.Cache.BUILD_TYPE) == build_type);
            cache = new CMake.Cache("../CMakeCache.txt");
            assert(cache.get_value(CMake.Cache.BUILD_TYPE) == build_type);
        });
        
        Test.add_func ("/euclide/cmake/file", () => {
            CMake.File file = new CMake.File("src/tests/main.vala", "/home/user/src/tests/main.vala");
            assert(file.relative == "src/tests/main.vala");
            assert(file.absolute == "/home/user/src/tests/main.vala");
            assert(file.relative_dir == "src/tests/");
            assert(file.name == "main.vala");
        });
        
        Test.add_func("/euclide/cmake/add_executable", test_cmake_add_executable);
        Test.add_func("/euclide/bazaar", test_bazaar);
    }

    void test_cmake_deps () {
        var parser = new CMake.Parser("../../src/CMakeLists.txt");
        assert(parser.project.lib[0].name == "euclide");
        assert(parser.project.lib[0].get_deps()[0].function_value == "euclidecore");

        assert(parser.project.lib[0].get_dependencies()[0] == "libxml-2.0");
        
        
        parser = new CMake.Parser("../../tests/test2/CMakeLists.txt");
        assert(parser.project.lib[0].name == "euclidecore");
        foreach(var dep in parser.project.lib[0].get_dependencies()) print("%s\n", dep);
    }

    void test_cmake_add_deps () {
        var parser = new CMake.Parser("../../tests/test1/CMakeLists.txt");
        assert(parser.project.bin[0].name == "test1");
        assert(parser.project.bin[0].get_deps()[0].function_value == "REQUIRED gtk+-3.0");
        parser.project.bin[0].add_dep("gee-1.0");
        parser.save_content();
        parser = new CMake.Parser("../../tests/test1/CMakeLists.txt");
        assert(parser.project.bin[0].get_deps()[0].function_value  == "REQUIRED gtk+-3.0 gee-1.0");
        parser.project.bin[0].remove_dep("gee-1.0");
        parser.save_content();
        parser = new CMake.Parser("../../tests/test1/CMakeLists.txt");
        assert(parser.project.bin[0].name == "test1");
        assert(parser.project.bin[0].get_deps()[0].function_value == "REQUIRED gtk+-3.0");
    }

    void test_cmake_add_file () {
        var parser = new CMake.Parser("../../tests/test2/CMakeLists.txt");
        assert(parser.project.bin[0].name == "euclide");
        assert(parser.project.lib[0].name == "euclidecore");
        parser.project.bin[0].add_file("test.vala");
        parser.project.lib[0].add_file("test_lib.vala");
        parser.save_content();

        parser = new CMake.Parser("../../tests/test2/CMakeLists.txt");
        assert(parser.project.bin[0].files[parser.project.bin[0].files.size - 2].relative == "test.vala");
        assert(parser.project.lib[0].files[1].relative == "test_lib.vala");
        parser.project.bin[0].remove_file("test.vala");
        parser.project.lib[0].remove_file("test_lib.vala");
        parser.save_content();
    }

    void test_cmake_add_executable()
    {        
        /* Set up env */
        try {
            Process.spawn_sync ("/", {"rm", "/tmp/test-add-exe", "-R"}, null, SpawnFlags.SEARCH_PATH, null);
        } catch (SpawnError e) {
            warning (e.message);
        }
        try {
            bool success = Process.spawn_sync (Environment.get_current_dir(),
                                               {"cp", "../../tests/test-add-exe/", "/tmp/", "-R"},
                                               null,
                                               SpawnFlags.SEARCH_PATH,
                                               null); /* FIXME: if you find a nicer way to do that… */
            assert(success == true);
        } catch (SpawnError e) {
            warning (e.message);
        }
        
        /* Tests */
        var parser = new CMake.Parser("/tmp/test-add-exe/CMakeLists.txt");
        assert(parser.project.bin.size == 1);
        
        assert(parser.add_target("new-executable") == true);
        assert(parser.project.bin.size == 2);
        assert(parser.add_target("new-executable") == false);
    }

    /**
     * Test for the new engine, started the 9/08/2011.
     **/
    void test_new_engine()
    {
        var cmd_list = new Gee.ArrayList<CMake.Command>();
        var command = CMake.Function.create_function("set(NAME \"euclide\")");
        assert(command  is CMake.VariableFunction);
        assert(command.vars_defined["NAME"] == "\"euclide\"");
        cmd_list.add(command);

        var command2 = CMake.Function.create_function("set(VERSION 0.1 ${NAME})");
        assert(command2.vars_defined["VERSION"] == "0.1 ${NAME}");
        assert(command2.vars_needed.has_key("NAME"));
        assert(command2.vars_needed["NAME"] == "");
        
        command2.resolve_vars(cmd_list);
        assert(command2.vars_needed["NAME"] == "\"euclide\"");
        assert(command2.vars_defined["VERSION"] == "0.1 \"euclide\"");
        
        
        /* With an extract of file */
        
        var commands = new Gee.ArrayList<CMake.Command>();
        CMake.BaseParser.parse_string(commands, 
            """
            set(NAME Euclide)
            set(LONG_NAME IDE ${NAME} 0.1)

            """);
        assert(commands.size == 2);
        assert(commands[1].vars_defined["LONG_NAME"] == "IDE Euclide 0.1");
        
        commands.clear();
        CMake.BaseParser.parse_string(commands, 
            """
            pkg_check_modules(DEPS gtk+-3.0 gio-2.0)
            """);
        assert(commands[0].vars_defined["DEPS_CFLAGS"] == "gtk+-3.0 gio-2.0");
        assert(commands[0].vars_defined["DEPS_LIBRARIES"] == "gtk+-3.0 gio-2.0");
        
        commands.clear();
        CMake.BaseParser.parse_string(commands, 
            """
            set(VAR_NAME test)
            pkg_check_modules(DEPS ${VAR_NAME} gtk+-3.0 gio-2.0)
            set(ANOTHER_VAR ${DEPS_CFLAGS} gee-1.0)
            """);
        assert(commands[0].vars_defined["VAR_NAME"] == "test");
        assert(commands[1].vars_defined["DEPS_CFLAGS"] == "test gtk+-3.0 gio-2.0");
        assert(commands[1].vars_defined["DEPS_LIBRARIES"] == "test gtk+-3.0 gio-2.0");
        assert(commands[2].vars_defined["ANOTHER_VAR"] == "test gtk+-3.0 gio-2.0 gee-1.0");
    }
    
    void test_new_engine_vala_precompile()
    {
        var parser = new CMake.CommandParser("vala_precompile(VALA_C
                                                main.vala
                                                ProjectManager.vala
                                                Settings.vala
                                            PACKAGES
                                                gtk+-3.0
                                                vte-2.90
                                            OPTIONS
                                                --thread
                                            )");
        CMake.ValaPrecompile vala_precompile = parser.create_command() as CMake.ValaPrecompile;
        assert(vala_precompile != null);
        assert(vala_precompile.variable_name == "VALA_C");
        assert(vala_precompile.files.nth_data(0) == "main.vala");
        assert(vala_precompile.options.nth_data(0) == "--thread");
        assert(vala_precompile.packages.nth_data(0) == "gtk+-3.0");
        
        vala_precompile.construct_command();
        assert(vala_precompile.command ==
            "vala_precompile(VALA_C\n    main.vala\n    ProjectManager.vala\n    Settings.vala\nPACKAGES\n    gtk+-3.0\n    vte-2.90\nOPTIONS\n    --thread\n)");
        
        /* We'll use a smaller function, just for readability. */ 
        parser = new CMake.CommandParser("vala_precompile(VALA_C main.vala PACKAGES gtk+-3.0)");
        vala_precompile = parser.create_command() as CMake.ValaPrecompile;
        assert(vala_precompile != null);
        vala_precompile.files.append("test.vala");
        vala_precompile.construct_command();
        assert(vala_precompile.command ==
            "vala_precompile(VALA_C\n    main.vala\n    test.vala\nPACKAGES\n    gtk+-3.0\nOPTIONS\n)");
        assert(vala_precompile.remove_file("test.vala"));
        vala_precompile.construct_command();
        assert(vala_precompile.command ==
            "vala_precompile(VALA_C\n    main.vala\nPACKAGES\n    gtk+-3.0\nOPTIONS\n)");
        
        
        /* Comments */
        parser = new CMake.CommandParser("vala_precompile(VALA_C
                                                main.vala
                                                ProjectManager.vala
                                                Settings.vala
                                                #test.vala
                                                # test2.vala
                                            PACKAGES
                                                gtk+-3.0
                                                vte-2.90
                                            OPTIONS
                                                --thread
                                            )");
        vala_precompile = parser.create_command() as CMake.ValaPrecompile;
        assert(vala_precompile.files.length() == 3);
    }
    
    void test_library_deps()
    {
        var parser = new CMake.Parser("../../tests/test_lib_deps/CMakeLists.txt");
        var project = parser.project;
        assert(project.name == "TestLibDeps");
        
        assert(project.lib.size == 1);
        var granite_lib = project.lib[0];
        assert(granite_lib.name == "granite");
        assert(granite_lib.related_functions.size == 3);
        var target_link = granite_lib.related_functions[2] as CMake.TargetLinkLibraries;
        assert(target_link != null);
        assert(target_link.vars_needed_cmd.size == 2);
        
        var granite_deps = granite_lib.get_dependencies();
        assert(granite_deps.size == 2);
        assert(granite_deps[0] == "gtk+-3.0");
        assert(granite_deps[1] == "gio-unix-2.0");
        
        granite_lib.add_dep("glib-2.0");
        parser.save_content();

        parser = new CMake.Parser("../../tests/test_lib_deps/CMakeLists.txt");
        project = parser.project;
        granite_lib = project.lib[0];
        
        granite_deps = granite_lib.get_dependencies();
        assert(granite_deps.size == 3);
        assert(granite_deps[0] == "gtk+-3.0");
        assert(granite_deps[1] == "gio-unix-2.0");
        assert(granite_deps[2] == "glib-2.0");
        
        granite_lib.remove_dep("glib-2.0");
        parser.save_content();
        
        parser = new CMake.Parser("../../tests/test_lib_deps/CMakeLists.txt");
        project = parser.project;
        granite_lib = project.lib[0];
        
        granite_deps = granite_lib.get_dependencies();
        assert(granite_deps.size == 2);
        assert(granite_deps[0] == "gtk+-3.0");
        assert(granite_deps[1] == "gio-unix-2.0");
    }
    
    void test_source_view()
    {
        /* Automatic indentation */
        
    }
    
    void test_bazaar()
    {
        assert(is_bzr_branch(".") == true);
        assert(is_bzr_branch("/tmp") == false);
    }
    
    void main(string[] args)
    {
        Gtk.init(ref args);
        Test.init(ref args);

        //add_cmake_tests ();
        Test.add_func("/euclide/cmake/add_dependencies", test_cmake_add_deps);
        Test.add_func("/euclide/newengine", test_new_engine);
        Test.add_func("/euclide/newengine_vala_precompile", test_new_engine_vala_precompile);
        Test.add_func("/euclide/cmake/add_file_complex_file", test_cmake_add_file);
        Test.add_func("/euclide/cmake/dependencies", test_cmake_deps);
        Test.add_func("/euclide/cmake/dependencies_lib", test_library_deps);

        Idle.add (() => {
            Test.run ();
            Gtk.main_quit ();
            return false;
            }
        );
        Gtk.main ();
    }
}