/*
 * Copyright (c) 2011 Lucas Baudin <xapantu@gmail.com>
 *
 * This is a free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; see the file COPYING.  If not,
 * write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

enum Euclide.Plugins.SymbolType
{
    NONE,
    FUNCTION,
    CLASS,
    NAMESPACE,
    STRUCT,
    ENUM
}
class Euclide.Plugins.FileParser : Object
{
    string content;
    public Gtk.TreeStore symbols;
    string file;
    public FileParser(string file, Gtk.TreeStore? store = null)
    {
        this.file = file;
        get_content();
        symbols = store ?? new Gtk.TreeStore(4, typeof(string), typeof(SymbolType), typeof(int), typeof(Gdk.Pixbuf));
        symbols.clear();
        parse_string(symbols, content);
    }
    
    void get_content()
    {
        FileUtils.get_contents(file, out content);
    }
    
    public static void parse_string(Gtk.TreeStore symbols, string content)
    {
        var lines = ("{\n" + content.replace("{{", "{ {").replace("}}", "} }") + "}\n").split("\n");
        int word_number = 0;
        bool update_word = false;
        bool commented = false;

        parse_block(symbols, null, 0, ref word_number, lines, 0);
    }
    
    static int string_char_count(string search, char to_search)
    {
        int count = 0;
        for(int i = 0; i < search.length; i++)
        {
            if(search[i] == to_search)
            {
                count ++;
            }
        }
        return count;
    }
    
    static string escape_string(string word, ref bool escaped)
    {
        string new_word = "";
        for(int i = 0; i < word.length; i++)
        {
            if(word[i] == '"')
            {
                escaped = !escaped;
            }
            if(!escaped)
            {
                new_word += word[i].to_string();
            }
        }
        return new_word;
    }
    
    public static int parse_block(Gtk.TreeStore symbols, Gtk.TreeIter? root_iter, int line_number, ref int word_number, string[] lines, int indent = 0, SymbolType parent = SymbolType.NONE, bool is_function = false)
    {
        int block_opened = 0;
        bool dont_reinit_word = false;
        string last_word = "";

        /* handle the differents controls */
        bool next_is_namespace = false;
        bool next_is_class = false;
        bool next_is_struct = false;
        bool next_is_enum = false;
        bool args_started = false;
        bool args_finished = false;
        string? function_candidate = "";
        bool commented = false;
        bool escaped = false;
        
        SymbolType current = parent;
        Gtk.TreeIter? iter = null;
        Gtk.TreeIter? class_iter = null;

        for(int i = line_number; i < lines.length; i++)
        {
            if(i == -1)
                break;
            string[] words = lines[i].split(" ");
            for(int j = word_number; j < words.length; j++)
            {
                string word = words[j];
                
                if(word.contains("/*"))
                {
                    commented = true;
                }
                if(word.contains("*/"))
                {
                    commented = false;
                }
                if(word.contains("\""))
                {
                    word = escape_string(word, ref escaped);
                }
                if(word != "" && !commented)
                {
                    
                    if(next_is_namespace)
                    {
                        next_is_namespace = false;
                        symbols.append(out iter, root_iter);
                        symbols.set(iter, 0, word, 1, SymbolType.NAMESPACE, 2, i, 3, pix_namespace, -1);
                        current = SymbolType.NAMESPACE;
                    }
                    if(next_is_class)
                    {
                        next_is_class = false;
                        symbols.append(out class_iter, iter ?? root_iter);
                        symbols.set(class_iter, 0, word, 1, SymbolType.CLASS, 2, i, 3, pix_class, -1);
                        current = SymbolType.CLASS;
                    }
                    if(next_is_enum)
                    {
                        Gtk.TreeIter? tree_iter = null;
                        next_is_enum = false;
                        symbols.append(out tree_iter, iter ?? root_iter);
                        symbols.set(tree_iter, 0, word, 1, SymbolType.ENUM, 2, i, 3, pix_enum, -1);
                        current = SymbolType.ENUM;
                    }
                    if(next_is_struct)
                    {
                        Gtk.TreeIter? tree_iter = null;
                        next_is_struct = false;
                        symbols.append(out tree_iter, iter ?? root_iter);
                        symbols.set(tree_iter, 0, word, 1, SymbolType.STRUCT, 2, i, 3, pix_struct, -1);
                        current = SymbolType.STRUCT;
                    }
                    
                    if(word.contains("(") && !is_function)
                    {
                        if(word[0] != '(')
                        {
                            function_candidate = word.split("(")[0];
                        }
                        else
                        {
                            function_candidate = last_word;
                        }
                        args_started = true;
                    }
                    
                    if(word.contains(")"))
                    {
                        args_finished = true;
                    }
                    
                    if(word.contains(";"))
                    {
                        function_candidate = null;
                        args_started = false;
                        args_finished = false;
                    }
                    
                    if(word.contains("}"))
                    {
                        //debug(indent + "} at line %d", i);
                        block_opened -= string_char_count(word, '}');
                        if(block_opened == 0 && ! word.contains("{")) /* the ! contains "{" is here to avoid returns with things like {} */
                        {
                            //debug(indent + "Main block closed.");
                            word_number = j;
                            return i;
                        }
                        else if(block_opened == 1)
                        {
                            //debug(indent + "Child block closed.");
                        }
                    }
                    if(word.contains("{"))
                    {
                        //debug(indent.to_string() + "{ at line %d", i);
                        block_opened += string_char_count(word, '{');
                        switch(block_opened)
                        {
                        case 0: /* should not happen */
                            //critical("Negative number of blocks.");
                            break;
                        case 1: /* main block opened */
                            //debug("%d Main block opened.", indent);
                            break;
                        case 2: /* child block opened */
                            //debug(indent + "Child block opened.");
                            if(!is_function) /* it is not necessary to parse the blocks which are in the functions */
                            {
                                bool is_block_function = args_started && args_finished && function_candidate != null  /* the last one shouldn't be necessary */;
                                if(is_block_function)
                                {
                                    Gtk.TreeIter? iter_func;
                                    symbols.append(out iter_func, class_iter ??  iter ?? root_iter);
                                    symbols.set(iter_func, 0, function_candidate, 1, SymbolType.FUNCTION, 2, i, 3, pix_function, -1);
                                }
                                i = parse_block(symbols, class_iter ?? iter, i, ref j, lines, indent + 1, current, is_block_function) - 1;
                                word_number = j;
                                function_candidate = null;
                                args_started = false;
                                args_finished = false;
                            }
                            break;
                        default:
                            if(!is_function)
                                critical("Child of a child? it shouldn't happen: %d", block_opened);
                            break;
                        }
                    }
                    if(block_opened == 2 && !is_function)
                    {
                        dont_reinit_word = true;
                        break;
                    }

                    check_keyword(word, ref next_is_namespace, ref next_is_class, ref next_is_enum, ref next_is_struct);
                    last_word = word;
                }
            }
            if(dont_reinit_word)
            {
                dont_reinit_word = false;
            }
            else
            {
                word_number = 0;
            }
        }
        return -1;
    }
    
    static void check_keyword(string word, ref bool is_namespace, ref bool is_class, ref bool is_enum, ref bool is_struct)
    {
        switch(word)
        {
        case "namespace":
            is_namespace = true;
            break;
        case "class":
            is_class = true;
            break;
        case "enum":
            is_enum = true;
            break;
        case "struct":
            is_struct = true;
            break;
        }
    }
}