/*
 * Copyright (C) 2008-2010 Abderrahim Kitouni
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

public class ValaProvider : Object, Gtk.SourceCompletionProvider {
	Gtk.TextIter start_pos;
	ValaPlugin? plugin;

	static Regex member_access;
	static Regex member_access_split;
	static Regex function_call;

	const string PREF_AUTOCOMPLETE_ENABLE = "cpp-completion-enable";
	const string PREF_SPACE_AFTER_FUNC = "cpp-completion-space-after-func";
	const string PREF_BRACE_AFTER_FUNC = "cpp-completion-brace-after-func";
	internal const string PREF_CALLTIP_ENABLE = "cpp-calltip-enable";

	static construct {
		try {
			member_access = new Regex("""((?:\w+(?:\s*\([^()]*\))?\.)*)(\w*)$""");
			member_access_split = new Regex ("""(\s*\([^()]*\))?\.""");
			function_call = new Regex("""(new )?((?:\w+(?:\s*\([^()]*\))?\.)*)(\w+)\s*\(([^(,)]+,)*([^(,)]*)$""");
		} catch(RegexError err) {
			critical("Regular expressions failed to compile : %s", err.message);
		}
	}

	public ValaProvider(ValaPlugin plugin) {
		this.plugin = plugin;
	}
	public string get_name () {
		return "Vala";
	}
	
	public void populate (Gtk.SourceCompletionContext context) {
		var proposals = new GLib.List<ValaProposal>();
		start_pos = plugin.interf.get_position ();
		var line_start = plugin.interf.get_line_begin_position (plugin.interf.get_lineno ());
		var current_text = plugin.interf.get_text(line_start, start_pos);

		MatchInfo match_info;
		if (! member_access.match(current_text, 0, out match_info))
			return;
		if (match_info.fetch(0).length < 2)
			return;

		start_pos.set_offset (start_pos.get_offset() - (int) match_info.fetch(2).length);

		var names = member_access_split.split (match_info.fetch(1));

		var syms = plugin.lookup_symbol (construct_member_access (names), match_info.fetch(2),
		                                 true, plugin.get_current_context () as Vala.Block);

		foreach (var symbol in syms) {
			if (symbol is Vala.LocalVariable
			    && symbol.source_reference.begin.line > plugin.interf.get_lineno())
				continue;

			proposals.prepend (new ValaProposal (symbol.name, symbol));
		}
		proposals.reverse();
		context.add_proposals (this, proposals, true);
	}
	public bool get_start_iter (Gtk.SourceCompletionContext context, Gtk.SourceCompletionProposal proposal, Gtk.TextIter iter) {
		iter = start_pos;
		return true;
	}
	public bool activate_proposal (Gtk.SourceCompletionProposal proposal, Gtk.TextIter iter) {
		var sym = (proposal as ValaProposal).data as Vala.Symbol;
		var editor = plugin.interf.editor;
		var assist = sym.name;
		var is_func = false;
		var calltip = false;

		if (sym is Vala.Method || sym is Vala.Signal) {
			is_func = true;
		} else if (sym is Vala.Variable) {
			if (((Vala.Variable) sym).variable_type is Vala.DelegateType) {
				is_func = true;
			}
		}

		if (is_func) {
			if (true) {//plugin.settings.get_boolean (PREF_SPACE_AFTER_FUNC)) {
				assist += " ";
			}
			if (true) {//plugin.settings.get_boolean (PREF_BRACE_AFTER_FUNC)) {
				assist += "(";
				if (true) {//plugin.settings.get_boolean (PREF_CALLTIP_ENABLE)) {
					calltip = true;
				}
			}
		}

		editor.buffer.delete(ref start_pos, ref iter);
		editor.buffer.insert(ref start_pos, assist, -1);

		if (calltip) {
			show_call_tip ();
		}
		
		return true;
	}

	public void show_call_tip () {
		var current_position = plugin.interf.get_position ();
		var line_start = plugin.interf.get_line_begin_position(plugin.interf.get_lineno());
		var to_complete = plugin.interf.get_text(line_start, current_position);

		List<string> tips = null;

		MatchInfo match_info;
		if (! function_call.match(to_complete, 0, out match_info))
			return;

		var creation_method = (match_info.fetch(1) != "");
		var names = member_access_split.split (match_info.fetch(2));
		var syms = plugin.lookup_symbol (construct_member_access (names), match_info.fetch(3),
		                                 false, plugin.get_current_context () as Vala.Block);

		foreach (var sym in syms) {
			Vala.List<Vala.Parameter> parameters = null;
			if (sym is Vala.Method) {
				parameters = ((Vala.Method) sym).get_parameters ();
			} else if (sym is Vala.Signal) {
				parameters = ((Vala.Signal) sym).get_parameters ();
			} else if (creation_method && sym is Vala.Class) {
				parameters = ((Vala.Class)sym).default_construction_method.get_parameters ();
			} else if (sym is Vala.Variable) {
				var var_type = ((Vala.Variable) sym).variable_type;
				if (var_type is Vala.DelegateType) {
					parameters = ((Vala.DelegateType) var_type).delegate_symbol.get_parameters ();
				} else {
					return;
				}
			} else {
				return;
			}
			var calltip = new StringBuilder ("(");
			var first = true;
			foreach (var p in parameters) {
				if(first) {
					first = false;
				} else {
					calltip.append(", ");
				}
				if (p.ellipsis) {
					calltip.append("...");
				} else {
					calltip.append (p.variable_type.to_qualified_string());
					calltip.append (" ").append (p.name);
				}
			}
			calltip.append (")");
			tips.append (calltip.str);
		}
		plugin.interf.show_tips (tips, plugin.interf.get_position ());
	}
	Vala.Expression construct_member_access (string[] names) {
		Vala.Expression expr = null;

		for (var i = 0; names[i] != null; i++) {
			if (names[i] != "") {
				expr = new Vala.MemberAccess (expr, names[i]);
				if (names[i+1] != null && names[i+1].chug ().has_prefix ("(")) {
					expr = new Vala.MethodCall (expr);
					i++;
				}
			}
		}

		return expr;
	}

}

