/*
 * Copyright (C) 2008-2010 Abderrahim Kitouni
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

using GLib;

public class ValaPlugin {
	
	internal ValaAnjutaInterface interf;

	Vala.CodeContext context;
	Cancellable cancel;
	BlockLocator locator = new BlockLocator ();

	AnjutaReport reporter;
	ValaProvider provider;

	Vala.Parser parser;

	Vala.Set<string> current_sources = new Vala.HashSet<string> (str_hash, str_equal);

	public ValaPlugin (ValaAnjutaInterface _interf) {

		interf = _interf;
		reporter = new AnjutaReport();
		reporter.docman = interf;
		parser = new Vala.Parser ();

		init_context ();

		provider = new ValaProvider(this);

		interf.char_added.connect (on_char_added);
		interf.saved.connect (on_file_saved);
	}

	void init_context () {
		context = new Vala.CodeContext();
		context.profile = Vala.Profile.GOBJECT;
		context.report = reporter;
		context.compile_only = true;
		reporter.clear_error_indicators ();

		cancel = new Cancellable ();

		/* This doesn't actually parse anything as there are no files yet,
		   it's just to set the context in the parsers */
		parser.parse (context);

		current_sources = new Vala.HashSet<string> (str_hash, str_equal);

	}

	void parse () {
		reporter.clear_error_indicators (null);
		try {
			Thread.create<void>(() => {
				lock (context) {
					Vala.CodeContext.push(context);
					var reporter = context.report as AnjutaReport;

					foreach (var src in context.get_source_files ()) {
						if (src.get_nodes ().size == 0) {
							debug ("parsing file %s", src.filename);
							parser.visit_source_file (src);
						}

						if (cancel.is_cancelled ()) {
							Vala.CodeContext.pop();
							return;
						}
					}

					if (reporter.get_errors () > 0 || cancel.is_cancelled ()) {
						Vala.CodeContext.pop();
						return;
					}

					context.check ();
					Vala.CodeContext.pop();
				}
			}, false);
		} catch (ThreadError err) {
			warning ("cannot create thread : %s", err.message);
		}
	}

	const string ATTRIBUTES = FileAttribute.STANDARD_NAME + "," + FileAttribute.STANDARD_TYPE;
	void open_dir_rec (File dir)
	{
		try {
			var enumerator = dir.enumerate_children (ATTRIBUTES, FileQueryInfoFlags.NOFOLLOW_SYMLINKS, null);
			FileInfo? file_info = null;
			
			while ((file_info = enumerator.next_file ()) != null) {
				var file_name = file_info.get_name ();
				var file_type = file_info.get_file_type ();
				
				if (file_type == FileType.REGULAR && (file_name.has_suffix (".vala") || file_name.has_suffix (".vapi"))) {
					var path = dir.get_child (file_name).get_path ();
					if (!(path in current_sources)) {
						context.add_source_filename (path);
						current_sources.add (path);
						print ("Added %s\n", path);
					}
				} else if (file_type == FileType.DIRECTORY) {
					open_dir_rec (dir.get_child (file_name));
				}
			}
		} catch (Error e) { warning (e.message); }
	}

	void add_project_files () {

		Vala.CodeContext.push (context);

		var src_dir = interf.project_dir.get_child ("src");
		if (src_dir.query_exists ())
			open_dir_rec (src_dir);

		var vapi_dir = interf.project_dir.get_child ("vapi");
		if (vapi_dir.query_exists ())
			open_dir_rec (vapi_dir);

		if (!context.has_package ("gobject-2.0")) {
			context.add_external_package("glib-2.0");
			context.add_external_package("gobject-2.0");
		}
		if (!context.has_package ("granite"))
			context.add_external_package ("granite");

		string[] flags = {};

		string[] packages = {};
		string[] vapidirs = {};

		for (int i = 0; i < flags.length; i++) {
			if (flags[i] == "--vapidir")
				vapidirs += flags[++i];
			else if (flags[i].has_prefix ("--vapidir="))
				vapidirs += flags[i].substring ("--vapidir=".length);
			else if (flags[i] == "--pkg")
				packages += flags[++i];
			else if (flags[i].has_prefix ("--pkg="))
				packages += flags[i].substring ("--pkg=".length);
			else
				debug ("Unknown valac flag %s", flags[i]);
		}

		context.vapi_directories = vapidirs;
		foreach (var pkg in packages) {
			if (context.has_package (pkg)) {
				debug ("package %s skipped", pkg);
			} else if (context.add_external_package(pkg)) {
				debug ("package %s added", pkg);
			} else {
				debug ("package %s not found", pkg);
			}
		}
		Vala.CodeContext.pop();
	}

	public bool add_package (string pkg)
	{
		if (context.has_package (pkg))
			return false;
		if (context.add_external_package (pkg)) {
			parse ();
			return true;
		}
		
		warning ("Package %s not found", pkg);
		return false;
	}

	/* "document_manager_current_document" watch */
	public bool set_current_document (File file) {

		if (!(file.get_path () in current_sources)) {
			cancel.cancel ();
			lock (context) {
				init_context ();
				add_project_files ();
			}

			parse ();
		}

		reporter.update_errors ();
		
		return current_sources.size > 0;
	}

	public void on_file_saved (File file) {
		foreach (var source_file in context.get_source_files ()) {
			if (source_file.filename != file.get_path())
				continue;

			uint8[] contents;
			try {
				file.load_contents (null, out contents, null);
				source_file.content = (string) contents;
				update_file (source_file);
			} catch (Error e) {
				warning (e.message);
			}
			return;
		}
	}

	public void on_char_added (Gtk.TextIter position, string ch) {

		if (ch == "(") {
			provider.show_call_tip ();
		} else if (ch == ")") {
			interf.hide_tip ();
		}
	}

	Vala.Symbol? lookup_symbol_by_cname (string cname, Vala.Symbol parent=context.root) {
		var sym = parent.scope.lookup (cname);
		if (sym != null)
			return sym;

		var symtab = parent.scope.get_symbol_table ();
		foreach (var name in symtab.get_keys ()) {
			if (cname.has_prefix (name)) {
				return lookup_symbol_by_cname (cname.substring (name.length), parent.scope.lookup (name));
			}
		}
		return null;
	}

	internal Vala.Symbol get_current_context (Gtk.TextIter? position = null) {

		var path = interf.file.get_path();
		
		lock (context) {
			Vala.SourceFile source = null;
			foreach (var src in context.get_source_files()) {
				if (src.filename == path) {
					source = src;
					break;
				}
			}
			if (source == null) {
				source = new Vala.SourceFile (context,
				                              path.has_suffix("vapi") ? Vala.SourceFileType.PACKAGE:
					                                                    Vala.SourceFileType.SOURCE,
				                              path);
				context.add_source_file(source);
				update_file(source);
			}
			int line; int column;
			if (position == null) {
				line = interf.get_lineno ();
				column = interf.get_column ();
			} else {
				line = position.get_line ();
				column = interf.get_line_begin_position (line).get_line_offset ();
			}
			return locator.locate(source, line, column);
		}
	}

	internal List<Vala.Symbol> lookup_symbol (Vala.Expression? inner, string name, bool prefix_match, Vala.Block block) {
		List<Vala.Symbol> matching_symbols = new List<Vala.Symbol> ();

		lock (context) {
			if (inner == null) {
				for (var sym = (Vala.Symbol) block; sym != null; sym = sym.parent_symbol) {
					matching_symbols.concat (symbol_lookup_inherited (sym, name, prefix_match));
				}

				foreach (var ns in block.source_reference.file.current_using_directives) {
					matching_symbols.concat (symbol_lookup_inherited (ns.namespace_symbol, name, prefix_match));
				}
			} else if (inner.symbol_reference != null) {
					matching_symbols.concat (symbol_lookup_inherited (inner.symbol_reference, name, prefix_match));
			} else if (inner is Vala.MemberAccess) {
				var inner_ma = (Vala.MemberAccess) inner;
				var matching = lookup_symbol (inner_ma.inner, inner_ma.member_name, false, block);
				if (matching != null)
					matching_symbols.concat (symbol_lookup_inherited (matching.data, name, prefix_match));
			} else if (inner is Vala.MethodCall) {
				var inner_inv = (Vala.MethodCall) inner;
				var inner_ma = inner_inv.call as Vala.MemberAccess;
				if (inner_ma != null) {
					var matching = lookup_symbol (inner_ma.inner, inner_ma.member_name, false, block);
					if (matching != null)
						matching_symbols.concat (symbol_lookup_inherited (matching.data, name, prefix_match, true));
				}
			}
		}
		return matching_symbols;
	}
	List<Vala.Symbol> symbol_lookup_inherited (Vala.Symbol? sym, string name, bool prefix_match, bool invocation = false) {
		List<Vala.Symbol> result = null;

		// This may happen if we cannot find all the needed packages
		if (sym == null)
			return result;

		var symbol_table = sym.scope.get_symbol_table ();
		if (symbol_table != null) {
			foreach (string key in symbol_table.get_keys()) {
				if (((prefix_match && key.has_prefix (name)) || key == name)) {
					result.append (symbol_table[key]);
				}
			}
		}
		if (invocation && sym is Vala.Method) {
			var func = (Vala.Method) sym;
			result.concat (symbol_lookup_inherited (func.return_type.data_type, name, prefix_match));
		} else if (sym is Vala.Class) {
			var cl = (Vala.Class) sym;
			foreach (var base_type in cl.get_base_types ()) {
				result.concat (symbol_lookup_inherited (base_type.data_type, name, prefix_match));
			}
		} else if (sym is Vala.Struct) {
			var st = (Vala.Struct) sym;
			result.concat (symbol_lookup_inherited (st.base_type.data_type, name, prefix_match));
		} else if (sym is Vala.Interface) {
			var iface = (Vala.Interface) sym;
			foreach (var prerequisite in iface.get_prerequisites ()) {
				result.concat (symbol_lookup_inherited (prerequisite.data_type, name, prefix_match));
			}
		} else if (sym is Vala.LocalVariable) {
			var variable = (Vala.LocalVariable) sym;
			result.concat (symbol_lookup_inherited (variable.variable_type.data_type, name, prefix_match));
		} else if (sym is Vala.Field) {
			var field = (Vala.Field) sym;
			result.concat (symbol_lookup_inherited (field.variable_type.data_type, name, prefix_match));
		} else if (sym is Vala.Property) {
			var prop = (Vala.Property) sym;
			result.concat (symbol_lookup_inherited (prop.property_type.data_type, name, prefix_match));
		} else if (sym is Vala.Parameter) {
			var fp = (Vala.Parameter) sym;
			result.concat (symbol_lookup_inherited (fp.variable_type.data_type, name, prefix_match));
		}

		return result;
	}
	void update_file (Vala.SourceFile file) {
		lock (context) {
			/* Removing nodes in the same loop causes problems (probably due to ReadOnlyList)*/
			var nodes = new Vala.ArrayList<Vala.CodeNode> ();
			foreach (var node in file.get_nodes()) {
				nodes.add(node);
			}
			foreach (var node in nodes) {
				file.remove_node (node);
				if (node is Vala.Symbol) {
					var sym = (Vala.Symbol) node;
					if (sym.owner != null)
						/* we need to remove it from the scope*/
						sym.owner.remove(sym.name);
					if (context.entry_point == sym)
						context.entry_point = null;
				}
			}
			file.current_using_directives = new Vala.ArrayList<Vala.UsingDirective>();
			var ns_ref = new Vala.UsingDirective (new Vala.UnresolvedSymbol (null, "GLib"));
			file.add_using_directive (ns_ref);
			context.root.add_using_directive (ns_ref);

			reporter.clear_error_indicators (file);

			parse ();

			reporter.update_errors();
		}
	}
}
